# from qt import *

import time

from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy


class PX1Energy(Device, AbstractEnergy):
    energy_state = {
        "ALARM": "error",
        "FAULT": "error",
        "RUNNING": "moving",
        "MOVING": "moving",
        "STANDBY": "ready",
        "DISABLE": "error",
        "UNKNOWN": "unknown",
        "EXTRACT": "outlimits",
    }

    def init(self):
        self.moving = False

        self.doBacklashCompensation = False

        self.current_energy = None
        self.current_state = None

        try:
            self.monodevice = DeviceProxy(self.get_property("mono_device"))
        except Exception:
            self.errorDeviceInstance(self.get_property("mono_device"))

        # Nom du device bivu (Energy to gap) : necessaire pour amelioration du
        # positionnement de l'onduleur (Backlash)
        self.und_device = DeviceProxy(self.get_property("undulator_device"))
        self.doBacklashCompensation = self.get_property("backlash")

        # parameters for polling
        self.is_connected()

        self.energy_chan = self.get_channel_object("energy")
        self.energy_chan.connect_signal("update", self.energyChanged)

        self.stop_cmd = self.get_command_object("stop")

        self.state_chan = self.get_channel_object("state")
        self.state_chan.connect_signal("update", self.stateChanged)

    def connect_notify(self, signal):
        if signal == "energyChanged":
            self.log.debug(
                "PX1Energy. connect_notify. sending energy value %s" % self.get_value()
            )
            self.energyChanged(self.get_energy())

        if signal == "stateChanged":
            self.log.debug(
                "PX1Energy. connect_notify. sending state value %s" % self.get_state()
            )
            self.stateChanged(self.get_state())

        self.set_is_ready(True)

    def stateChanged(self, value):
        str_state = str(value)
        if str_state == "MOVING":
            self.moveEnergyCmdStarted()

        if self.current_state == "MOVING" or self.moving == True:
            if str_state != "MOVING":
                self.moveEnergyCmdFinished()

        self.current_state = str_state
        self.emit("stateChanged", self.energy_state[str_state])

    # function called during polling
    def energyChanged(self, value):
        if (
            self.current_energy is not None
            and abs(self.current_energy - value) < 0.0001
        ):
            return

        self.current_energy = value

        wav = self.get_wavelength()
        if wav is not None:
            self.emit("energyChanged", (value, wav))

    def isSpecConnected(self):
        return True

    def is_connected(self):
        return True

    def sConnected(self):
        self.emit("connected", ())

    def sDisconnected(self):
        self.emit("disconnected", ())

    def isDisconnected(self):
        return True

    def get_value(self):
        return self.energy_chan.get_value()

    def get_state(self):
        return str(self.state_chan.get_value())

    def getEnergyComputedFromCurrentGap(self):
        return self.und_device.energy

    def getCurrentUndulatorGap(self):
        return self.und_device.gap

    def get_wavelength(self):
        return self.monodevice.read_attribute("lambda").value

    def get_wavelength(self):
        return self.get_wavelength()

    def get_limits(self):
        chan_info = self.energy_chan.getInfo()
        return (float(chan_info.min_value), float(chan_info.max_value))

    def get_wavelength_limits(self):
        energy_min, energy_max = self.get_limits()

        # max is min and min is max
        max_lambda = self.energy_to_lambda(energy_min)
        min_lambda = self.energy_to_lambda(energy_max)

        return (min_lambda, max_lambda)

    def energy_to_lambda(self, value):
        # conversion is done by mono device
        self.monodevice.simEnergy = value
        return self.monodevice.simLambda

    def lambda_to_energy(self, value):
        # conversion is done by mono device
        self.monodevice.simLambda = value
        return self.monodevice.simEnergy

    def set_value(self, value, wait=False):
        value = float(value)

        backlash = 0.1  # en mm
        gaplimite = 5.5  # en mm

        if self.get_state() != "MOVING":
            if self.doBacklashCompensation:
                try:
                    # Recuperation de la valeur de gap correspondant a l'energie souhaitee
                    # self.und_device.autoApplyComputedParameters = False
                    self.und_device.energy = value
                    newgap = self.und_device.computedGap
                    actualgap = self.und_device.gap

                    #                    self.und_device.autoApplyComputedParameters = True

                    while str(self.und_device.State()) == "MOVING":
                        time.sleep(0.2)

                    # On applique le backlash que si on doit descendre en gap
                    if newgap < actualgap + backlash:
                        # Envoi a un gap juste en dessous (backlash)
                        if newgap - backlash > gaplimite:
                            self.und_device.gap = newgap - backlash
                            while str(self.und_device.State()) == "MOVING":
                                time.sleep(0.2)

                            self.energy_chan.set_value(value)
                        else:
                            self.und_device.gap = gaplimite
                            self.und_device.gap = newgap + backlash
                        time.sleep(1)
                except Exception:
                    self.log.error(
                        "%s: Cannot move undulator U20 : State device = %s",
                        self.id,
                        str(self.und_device.State()),
                    )

            try:
                self.energy_chan.set_value(value)
                return value
            except Exception:
                self.log.error(
                    "%s: Cannot move Energy : State device = %s",
                    self.id,
                    self.get_state(),
                )

        else:
            self.log.error(
                "%s: Cannot move Energy : State device = %s",
                self.id,
                self.get_state(),
            )

    def set_wavelength(self, value, wait=False):
        egy_value = self.lambda_to_energy(float(value))
        self.log.debug(
            "%s: Moving wavelength to : %s (egy to %s" % (self.id, value, egy_value)
        )
        self.set_valuey(egy_value)
        return value

    def cancelMoveEnergy(self):
        self.stop_cmd()
        self.moving = False

    def energyLimitsChanged(self, limits):
        egy_min, egy_max = limits

        lambda_min = self.energy_to_lambda(egy_min)
        lambda_max = self.energy_to_lambda(egy_max)

        wav_limits = (lambda_min, lambda_max)

        self.emit("energyLimitsChanged", (limits,))

        if None not in wav_limits:
            self.emit("wavelengthLimitsChanged", (wav_limits,))
        else:
            self.emit("wavelengthLimitsChanged", (None,))

    def moveEnergyCmdReady(self):
        if not self.moving:
            self.emit("moveEnergyReady", (True,))

    def moveEnergyCmdNotReady(self):
        if not self.moving:
            self.emit("moveEnergyReady", (False,))

    def moveEnergyCmdStarted(self):
        self.moving = True
        self.emit("moveEnergyStarted", ())

    def moveEnergyCmdFailed(self):
        self.moving = False
        self.emit("moveEnergyFailed", ())

    def moveEnergyCmdAborted(self):
        self.moving = False

    def moveEnergyCmdFinished(self):
        self.moving = False
        self.emit("moveEnergyFinished", ())

    def getPreviousResolution(self):
        return (None, None)

    def restoreResolution(self):
        return (False, "Resolution motor not defined")


def test_hwo(hwo):
    print(hwo.get_value())
    print(hwo.get_wavelength())
    print(hwo.get_limits())
    print(hwo.getCurrentUndulatorGap())



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
BeamMockup class - methods to define the size and shape of he beam.

Example xml configuration:

.. code-block:: xml

  <object class="BeamMockup">
    <object href="/beam_definer" role="definer"/>
    <!-- accepted definer_type values: aperture, slits, definer -->
    <definer_type>definer</definer_type>
    <beam_divergence_vertical>0</beam_divergence_vertical>
    <beam_divergence_horizontal>0</beam_divergence_horizontal>
    <check_beam>(0.03, 0.03)</check_beam>
  </object>
"""

__copyright__ = """ Copyright © by MXCuBE Collaboration """
__license__ = "LGPLv3+"

from ast import literal_eval

from mxcubecore.HardwareObjects.abstract.AbstractBeam import AbstractBeam, BeamShape


class BeamMockup(AbstractBeam):
    """Beam Mockup class"""

    def __init__(self, name):
        super().__init__(name)
        self._definer_type = None
        self._check_beam = ()

    def init(self):
        """Initialize hardware"""
        super().init()

        #
        # backward compatibility hack to support loading from XML config file
        #
        # when loading from YAML configuration file,
        # the attributes will be automatically set to the specified child HWOBJs
        #
        # when loading from XML, it does not happen, so fall back to
        # get_object_by_role()
        #
        _definer_type = None
        self._aperture = self.get_object_by_role("aperture")
        self._slits = self.get_object_by_role("slits")
        self._definer = self.get_object_by_role("definer")

        if self.aperture:
            _definer_type = "aperture"
            self.aperture.connect("valueChanged", self.aperture_diameter_changed)

        if self.slits:
            _definer_type = "slits"
            self.slits.connect("valueChanged", self.slits_gap_changed)

        if self.definer:
            _definer_type = "definer"
            self.definer.connect("valueChanged", self._re_emit_values)

        self._definer_type = self.get_property("definer_type", _definer_type)

        self._beam_position_on_screen = self.get_property("beam_position") or [318, 238]
        if isinstance(self._beam_position_on_screen, str):
            self._beam_position_on_screen = literal_eval(self._beam_position_on_screen)

        self._check_beam = self.get_property("check_beam") or False
        if isinstance(self._check_beam, str):
            self._check_beam = literal_eval(self._check_beam)

        # Needed to trigger first value setting
        self.get_value()

        self.re_emit_values()
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def _re_emit_values(self, *args, **kwargs):
        self.re_emit_values()

    def _get_aperture_value(self) -> tuple[list[float, float], str]:
        """Get the size and the label of the aperture in place.

        Returns:
            Size [mm] (width, height), label.
        """
        _size = self.aperture.get_value().value[0]
        try:
            _label = self.aperture.get_value().name
        except AttributeError:
            _label = str(_size)
        _size /= 1000.0

        return [_size, _size], _label

    def _get_definer_value(self) -> tuple[list[float, float], str]:
        """Get the size and the name of the definer in place.

        Returns:
            Size [mm] (width, height), label.
        """
        try:
            value = self.definer.get_value()
            if isinstance(value, tuple):
                return [value[1], value[1]], value[0]
            return list(value.value), value.name
        except AttributeError:
            return [-1, -1], "UNKNOWN"

    def _get_slits_value(self) -> tuple[list[float, float], str]:
        """Get the size of the slits in place.

        Returns:
             Size [mm] (width, height), label.
        """
        _size = self.slits.get_gaps()
        return _size, "slits"

    def _get_value(self) -> tuple[float, float, BeamShape, str]:
        """Get the size (width and height) of the beam, its shape and
        its label. The size is in mm.

        Returns:
            Four-item tuple: width, height, shape, name
        """
        labels = {}
        _label = "UNKNOWN"
        if self.aperture:
            _size, _name = self._get_aperture_value()
            self._beam_size_dict.update({"aperture": _size})
            labels.update({"aperture": _name})

        if self.slits:
            _size, _name = self._get_slits_value()
            self._beam_size_dict.update({"slits": _size})
            labels.update({"slits": _name})

        if self.definer:
            _size, _name = self._get_definer_value()
            self._beam_size_dict.update({"definer": _size})
            labels.update({"definer": _name})

        info_dict = self.evaluate_beam_info()

        try:
            _label = labels[info_dict["label"]]
            self._beam_info_dict["label"] = _label
        except KeyError:
            _label = info_dict["label"]

        return self._beam_width, self._beam_height, self._beam_shape, _label

    def aperture_diameter_changed(self, aperture):
        """Method called when the aperture diameter changes.

        Args:
            Aperture enum.
        """

        size = aperture.value[0]
        self.aperture.update_value(aperture)
        self._beam_size_dict["aperture"] = [size, size]
        self.evaluate_beam_info()
        self._beam_info_dict["label"] = aperture.name
        self.re_emit_values()

    def slits_gap_changed(self, size: tuple[float, float]):
        """Method called when the slits gap changes.

        Args:
            Two floats - beam size in microns
        """
        self._beam_size_dict["slits"] = size
        self._beam_info_dict["label"] = "slits"
        self.evaluate_beam_info()
        self.re_emit_values()

    def set_beam_position_on_screen(self, beam_x_y: list[int, int]):
        """Sets beam mark position on screen.
        #TODO move method to sample_view

        Args:
            Position [x, y] in pixels.
        """
        self._beam_position_on_screen = beam_x_y
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def get_slits_gap(self) -> tuple[float, float]:
        """Get the beam size from the slits gap.

        Returns:
            Two-item tuple with horizontal and vertical beam size in microns
        """
        self.evaluate_beam_info()
        return self._beam_size_dict["slits"]

    def set_slits_gap(self, width_microns: int, height_microns: int):
        """Sets slits gap in microns.

        Args:
            width and height in microns.
        """
        if self.slits:
            self.slits.set_horizontal_gap(width_microns / 1000.0)
            self.slits.set_vertical_gap(height_microns / 1000.0)

    def get_aperture_pos_name(self) -> str:
        """Get the name of the current aperture.

        Returns:
             name of current aperture position
        """
        return self.aperture.get_current_pos_name()

    def get_defined_beam_size(self) -> dict:
        """Get the predefined beam labels and size.

        Returns:
            Dictionary with lists of available beam size labels
            and the corresponding size (width,height) tuples.
            ``{"label": [str, str, ...], "size": [(w,h), (w,h), ...]}``
        """
        labels = []
        values = []
        if self._definer_type == "slits":
            return {
                "label": ["low", "high"],
                "size": [self.slits.get_min_limits(), self.slits.get_max_limits()],
            }

        if self._definer_type == "aperture":
            _enum = self.aperture.VALUES
        elif self._definer_type == "definer":
            _enum = self.definer.VALUES

        for value in _enum:
            _nam = value.name
            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                labels.append(_nam)
                if self._definer_type == "aperture":
                    values.append((value.value[0] / 1000.0, value.value[0] / 1000.0))
                else:
                    values.append(value.value)
        return {"label": labels, "size": values}

    def get_available_size(self) -> dict:
        """Get the available predefined beam definer configuration.

        Returns:
            ``{"type": ["aperture"], "values": [labels]}`` or
            ``{"type": ["definer"], "values": [labels]}`` or
            ``{"type": ["width", "height"], "values":
                       [low_lim_w, high_lim_w, low_lim_h, high_lim_h]}``
        """
        if self._definer_type == "aperture":
            # get list of the available apertures
            return {
                "type": ["aperture"],
                "values": self.aperture.get_diameter_size_list(),
            }

        if self._definer_type == "definer":
            # get list of the available definer positions
            return {
                "type": ["definer"],
                "values": self.definer.get_predefined_positions_list(),
            }

        if self._definer_type == "slits":
            # get the list of the slits motors range
            _low_w, _low_h = self.slits.get_min_limits()
            _high_w, _high_h = self.slits.get_max_limits()
            return {
                "type": ["width", "height"],
                "values": [_low_w, _high_w, _low_h, _high_h],
            }

        return {}

    def set_value(self, size: list[float, float] | str | None = None):
        """Set the beam size.

        Args:
            size: List of width and  height in micrometers or
                  Aperture or definer definer name as string.
        Raises:
            RuntimeError: Beam definer not configured
                          Size out of the limits.
            TypeError: Wrong size type.
        """
        msg = "Incorrect input value for "
        if self._definer_type in (self.slits, "slits"):
            if not isinstance(size, list):
                msg += "slits"
                raise TypeError(msg)
            self.slits.set_horizontal_gap(size[0])
            self.slits.set_vertical_gap(size[1])

        if self._definer_type in (self.aperture, "aperture"):
            if not isinstance(size, str):
                msg += "aperture"
                raise TypeError(msg)
            self.aperture.set_value(self.aperture.VALUES[size], timeout=2)

        if self._definer_type in (self.definer, "definer"):
            if not isinstance(size, str):
                msg += "definer"
                raise TypeError(msg)
            self.definer.set_value(self.definer.VALUES[size], timeout=2)

    def _is_beam(self) -> bool:
        """Check if there is beam.

        Returns:
            ``True`` if beam present, ``False`` otherwise
        """
        if not self._check_beam:
            return True

        beam = self.get_value()
        return all(
            x1 <= x2
            for (x1, x2) in zip(self._check_beam, (beam[0], beam[1]), strict=False)
        )



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Enumerables and other constants used by the queue model.
"""

from collections import namedtuple

StrategyComplexity = namedtuple("StrategyComplexity", ["SINGLE", "FEW", "MANY"])
STRATEGY_COMPLEXITY = StrategyComplexity("none", "min", "full")

ExperimentType = namedtuple(
    "ExperimentType",
    [
        "SAD",
        "SAD_INV",
        "MAD",
        "MAD_INV",
        "NATIVE",
        "HELICAL",
        "EDNA_REF",
        "OSC",
        "MESH",
        "COLLECT_MULTIWEDGE",
        "STILL",
        "IMAGING",
    ],
)
EXPERIMENT_TYPE = ExperimentType(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
EXPERIMENT_TYPE_STR = ExperimentType(
    "SAD",
    "SAD - Inverse Beam",
    "MAD",
    "MAD - Inverse Beam",
    "OSC",
    "Helical",
    "Characterization",
    "OSC",
    "Mesh",
    "Collect - Multiwedge",
    "Still",
    "Imaging",
)

StrategyOption = namedtuple("StrategyOption", ["AVG"])
STRATEGY_OPTION = StrategyOption(0)

CollectionOrigin = namedtuple("CollectionOrigin", ["MXCUBE", "EDNA", "WORKFLOW"])
COLLECTION_ORIGIN = CollectionOrigin(0, 1, 2)
COLLECTION_ORIGIN_STR = CollectionOrigin("mxcube", "edna", "workflow")

EDNARefImages = namedtuple("EDNARefImages", ["FOUR", "TWO", "ONE", "NONE"])
EDNA_NUM_REF_IMAGES = EDNARefImages(0, 1, 2, 3)

CentringMethod = namedtuple(
    "CentringMethod", ["MANUAL", "LOOP", "FULLY_AUTOMATIC", "XRAY", "NONE"]
)
CENTRING_METHOD = CentringMethod(0, 1, 2, 3, 4)

WorkflowType = namedtuple(
    "WorkflowType", ["BURN", "WF1", "WF2", "LineScan", "MeshScan", "XrayCentring"]
)
WORKFLOW_TYPE = WorkflowType(0, 1, 2, 3, 4, 5)



from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import BeamInfo

"""
XML example file
<object class="ESRF.ESRFBeamInfo">
  <defaultBeamDivergence></defaultBeamDivergence>
  <object role="camera" hwrid="/prosilica_md2"/>
  <object role="aperture" hwrid="/udiff_aperturemot"/>
  <object role="diffractometer" hwrid="/udiff" />
  <!-- Positions and slits format: X Y -->
  <beam_position>322 243</beam_position>
  <beam_size_slits>0.04 0.04</beam_size_slits>
  <beam_divergence_vertical>6.5</beam_divergence_vertical>
  <beam_divergence_horizontal>104</beam_divergence_horizontal>
</object>
"""


class ESRFBeamInfo(BeamInfo.BeamInfo):
    def __init__(self, *args):
        BeamInfo.BeamInfo.__init__(self, *args)
        self.beam_position = (0, 0)

    def init(self):
        self.chan_beam_size_microns = None
        self.chan_beam_shape_ellipse = None
        BeamInfo.BeamInfo.init(self)

        beam_size_slits = self.get_property("beam_size_slits")
        if beam_size_slits:
            self.beam_size_slits = tuple(map(float, beam_size_slits.split()))

        beam_position = self.get_property("beam_position")
        if beam_position:
            self.beam_position = tuple(map(float, beam_position.split()))
        else:
            self.log.warning("ESRFBeamInfo: " + "beam position not configured")
        self.difrractometer_hwobj = self.get_object_by_role("difrractometer")
        self.flux = self.get_object_by_role("flux")
        self.beam_definer = self.get_object_by_role("beam_definer")

    def get_beam_position(self):
        if self.beam_position == (0, 0):
            try:
                self.beam_position = HWR.beamline.diffractometer.beam.get_value()
            except AttributeError:
                self.beam_position = (
                    HWR.beamline.sample_view.camera.get_width() / 2,
                    HWR.beamline.sample_view.camera.get_height() / 2,
                )
        return self.beam_position

    def set_beam_position(self, beam_x, beam_y):
        return

    def evaluate_beam_info(self, *args):
        BeamInfo.BeamInfo.evaluate_beam_info(self, *args)
        self.beam_info_dict["shape"] = "ellipse"
        return self.beam_info_dict



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import json
import os
import time

import gevent
import requests

from mxcubecore.CommandContainer import CommandObject

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class BlueskyHttpServerCommand(CommandObject):
    """Interface for communicating with the Bluesky Http Server API."""

    _default_timeout = 5
    _status_path = "api/status"
    _execute_path = "api/queue/item/execute"

    def __init__(self, name, url, timeout=5, **kwargs):
        CommandObject.__init__(self, name, **kwargs)
        self.username = ""
        self._default_timeout = timeout
        self._url = f"{url}/" if url[-1] != "/" else url
        self._headers = {"Authorization": f"ApiKey {os.environ['AUTH_KEY']}"}

    def format_response(self, response):
        if response:
            response.raise_for_status()
            return json.loads(response.text)
        return response

    def status(self):
        response = requests.get(
            self._url + self._status_path,
            headers=self._headers,
            timeout=self._default_timeout,
        )
        return self.format_response(response)

    def monitor_manager_state(self, stop_state, timeout=86400):
        with gevent.Timeout(timeout, exception=TimeoutError):
            while self.status()["manager_state"] != stop_state:
                time.sleep(0.1)

    def execute_plan(self, plan_name, kwargs=None):
        if not kwargs:
            kwargs = {}
        return requests.post(
            self._url + self._execute_path,
            headers=self._headers,
            json={
                "user": self.username,
                "item": {"name": plan_name, "item_type": "plan", "kwargs": kwargs},
            },
            timeout=self._default_timeout,
        )

    def is_connected(self):
        http_server_status = self.status()
        re_environment_open = http_server_status["worker_environment_exists"]
        re_running = http_server_status["re_state"] is not None
        return re_environment_open and re_running



"""
This module serves to connect to and Ldap server.

It works in principle for ESRF, Soleil Proxima and MAXIV beamlines
"""

import ldap

from mxcubecore.HardwareObjects.abstract.AbstractAuthenticator import (
    AbstractAuthenticator,
)

"""
ldapou is optional, if ldapou is not defined,
the bind_str (simple_bind) will be "uid=xxx,dc=xx,dc=xx",
otherwise it is uid=xxx,ou=xxx,dc=xx,dc=xx

<procedure class="LdapAuthenticator">
  <ldaphost>ldaphost.mydomain</ldaphost>
  <ldapport>389</ldapport>
  <ldapdomain>example.com</ldapdomain>
  <ldapou>users</ldapou>
</procedure>
"""


class LdapAuthenticator(AbstractAuthenticator):
    def __init__(self, name):
        super().__init__(name)
        self._ldapConnection = None

    # Initializes the hardware object
    def init(self):
        self._field_values = None
        self._connect()

    def _connect(self):
        ldaphost = self.get_property("ldaphost")
        ldapport = self.get_property("ldapport")
        domain = self.get_property("ldapdomain")

        if ldaphost is None:
            self.log.error("LdapAuthenticator: you must specify the LDAP hostname")
        else:
            if ldapport is None:
                self.log.debug(
                    "LdapAuthenticator: connecting to LDAP server %s", ldaphost
                )
                self._ldapConnection = ldap.initialize("ldap://" + ldaphost)
            else:
                self.log.debug(
                    "LdapAuthenticator: connecting to LDAP server %s:%s",
                    ldaphost,
                    ldapport,
                )
                self._ldapConnection = ldap.initialize(
                    "ldap://%s:%s" % (ldaphost, int(ldapport))
                )

            self.log.debug(
                "LdapAuthenticator: got connection %s" % str(self._ldapConnection)
            )

        if domain is not None:
            domparts = domain.split(".")
            domstr = ""
            comma = ""
            for part in domparts:
                domstr += "%sdc=%s" % (comma, part)
                comma = ","
            self.domstr = domstr
            self.log.debug(
                "LdapAuthenticator: got connection %s" % str(self._ldapConnection)
            )
        else:
            self.domstr = "dc=esrf,dc=fr"  # default is esrf.fr

    # Creates a new connection to LDAP if there's an exception on the current connection
    def _reconnect(self):
        if self._ldapConnection is not None:
            try:
                self._ldapConnection.result(timeout=0)
            except ldap.LDAPError:
                ldaphost = self.get_property("ldaphost")
                ldapport = self.get_property("ldapport")
                if ldapport is None:
                    self.log.debug(
                        "LdapAuthenticator: reconnecting to LDAP server %s", ldaphost
                    )
                    self._connect()
                else:
                    self.log.debug(
                        "LdapAuthenticator: reconnecting to LDAP server %s:%s",
                        ldaphost,
                        ldapport,
                    )
                    self._connect()

    def _cleanup(self, ex=None, msg=None):
        if ex is not None:
            try:
                msg = ex[0]["desc"]
            except (IndexError, KeyError, ValueError, TypeError):
                msg = "generic LDAP error"

            self._reconnect()

        self.log.info("LdapAuthenticator: %s" % msg)

        return False

    def get_field_values(self):
        return self._field_values

    def invalidate(self):
        pass

    def authenticate(self, username, password, retry=True, fields=None):
        # fields can be used in local implementation to retrieve user information from
        # ldap.  In ALBA for example, it is used to obtain homeDirectory upon successful
        # login and use that value for programming session hwo directories

        self._field_values = None

        if self._ldapConnection is None:
            return self._cleanup(msg="no LDAP server configured")

        self.log.debug(
            "LdapAuthenticator: searching for %s / %s" % (username, self.domstr)
        )
        try:
            search_str = self.domstr
            if fields is None:
                found = self._ldapConnection.search_s(
                    search_str, ldap.SCOPE_SUBTREE, "uid=" + username, ["uid"]
                )
            else:
                found = self._ldapConnection.search_s(
                    search_str, ldap.SCOPE_SUBTREE, "uid=" + username, fields
                )
        except ldap.LDAPError as err:
            if retry:
                self._cleanup(ex=err)
                return self.authenticate(username, password, retry=False)
            else:
                return self._cleanup(ex=err)

        if not found:
            return self._cleanup(msg="unknown proposal %s" % username)

        if fields is not None:
            self._field_values = found[0][1]

        if password == "":
            return self._cleanup(msg="invalid password for %s" % username)

        self.log.debug("LdapAuthenticator: validating %s" % username)

        try:
            bind_str = "uid=%s, ou=%s, %s" % (username, self.ldapou, self.domstr)
        except AttributeError:
            bind_str = "uid=%s,%s" % (username, self.domstr)

        self.log.debug("LdapAuthenticator: binding to %s" % bind_str)
        handle = self._ldapConnection.simple_bind(bind_str, password)

        try:
            self._ldapConnection.result(handle)
        except ldap.INVALID_CREDENTIALS:
            # try second time with different bind_str
            bind_str = "uid=%s, ou=people,%s" % (username, self.domstr)
            handle = self._ldapConnection.simple_bind(bind_str, password)
            try:
                self._ldapConnection.result(handle)
            except Exception:
                return self._cleanup(msg="invalid password for %s" % username)
        except ldap.LDAPError as err:
            if retry:
                self._cleanup(ex=err)
                return self.authenticate(username, password, retry=False)
            else:
                return self._cleanup(ex=err)

        return True



from tango import DevFailed

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import TangoChannel


class TangoAttributeReadError(Exception):
    def __init__(self, attribute_name: str):
        super().__init__()
        self.attribute_name = attribute_name


def add_attribute_channel(
    hwo: HardwareObject,
    tango_device: str,
    attribute_name: str,
    polling: int | None = None,
    update_callback=None,
) -> TangoChannel:
    """Utility function to add Tango attribute Channel to a hardware object.

    This function adds a Channel object to specified hardware object and checks
    that it's possible to read the specified tango attribute.

    If there is errors reading the attribute, the TangoAttributeReadError
    exception is raised.

    Parameters:
        hwo: The hardware object where to add the Channel object.
        tango_device: Tango device name.
        attribute_name: The tango attribute name.
        polling: Attribute polling periodicity.
        update_callback: Optional callback, if provided it is connected to
           "update" signal of Channel object
    """

    channel = hwo.add_channel(
        {
            "type": "tango",
            "tangoname": tango_device,
            "name": attribute_name,
            "polling": polling,
        },
        attribute_name,
    )

    if not channel.is_connected():
        raise TangoAttributeReadError(attribute_name)

    #
    # check if it's possible to read the Attribute
    #
    try:
        val = channel.get_value()
    except DevFailed:
        raise TangoAttributeReadError(attribute_name) from None

    #
    # add "update" callback
    #
    if update_callback is not None:
        channel.connect_signal("update", update_callback)
        #
        # We have 'consumed' the initial value of this channel above,
        # when we were testing if it's readable. The standard polling
        # will not invoke the callback with that value now.
        # Make sure initial value is passed to the callback.
        #
        update_callback(val)

    return channel



import logging
import os
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMultiCollect import (
    AbstractMultiCollect,
)
from mxcubecore.TaskUtils import task


class MultiCollectMockup(AbstractMultiCollect, HardwareObject):
    def __init__(self, name):
        AbstractMultiCollect.__init__(self)
        HardwareObject.__init__(self, name)
        self._centring_status = None
        self.ready_event = None
        self.actual_frame_num = 0

    def execute_command(self, command_name, *args, **kwargs):
        return

    def init(self):
        self.setControlObjects(
            diffractometer=HWR.beamline.diffractometer,
            sample_changer=HWR.beamline.sample_changer,
            lims=HWR.beamline.lims,
            safety_shutter=HWR.beamline.safety_shutter,
            machine_current=HWR.beamline.machine_info,
            cryo_stream=HWR.beamline.cryo,
            energy=HWR.beamline.energy,
            resolution=HWR.beamline.resolution,
            detector_distance=HWR.beamline.detector.distance,
            transmission=HWR.beamline.transmission,
            undulators=HWR.beamline.config.undulators,
            flux=HWR.beamline.flux,
            detector=HWR.beamline.detector,
            beam_info=HWR.beamline.beam,
        )
        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    @task
    def loop(self, owner, data_collect_parameters_list):
        failed_msg = "Data collection failed!"
        failed = True
        collections_analyse_params = []
        self.emit("collectReady", (False,))
        self.emit("collectStarted", (owner, 1))

        for data_collect_parameters in data_collect_parameters_list:
            logging.debug("collect parameters = %r", data_collect_parameters)
            failed = False
            data_collect_parameters["status"] = "Data collection successful"
            (
                osc_id,
                sample_id,
                sample_code,
                sample_location,
            ) = self.update_oscillations_history(data_collect_parameters)
            self.emit(
                "collectOscillationStarted",
                (
                    owner,
                    sample_id,
                    sample_code,
                    sample_location,
                    data_collect_parameters,
                    osc_id,
                ),
            )

            for image in range(
                data_collect_parameters["oscillation_sequence"][0]["number_of_images"]
            ):
                time.sleep(
                    data_collect_parameters["oscillation_sequence"][0]["exposure_time"]
                )
                self.emit("collectImageTaken", image)

            data_collect_parameters["status"] = "Running"
            data_collect_parameters["status"] = "Data collection successful"
            self.emit(
                "collectOscillationFinished",
                (
                    owner,
                    True,
                    data_collect_parameters["status"],
                    "12345",
                    osc_id,
                    data_collect_parameters,
                ),
            )

        self.emit(
            "collectEnded",
            owner,
            not failed,
            failed_msg if failed else "Data collection successful",
        )
        self.log.info("data collection successful in loop")
        self.emit("collectReady", (True,))

    @task
    def data_collection_hook(self, data_collect_parameters):
        return

    def do_prepare_oscillation(self, start, end, exptime, npass):
        self.actual_frame_num = 0

    @task
    def oscil(self, start, end, exptime, npass):
        return

    @task
    def data_collection_cleanup(self):
        return

    @task
    def close_fast_shutter(self):
        return

    @task
    def open_fast_shutter(self):
        return

    @task
    def move_motors(self, motor_position_dict):
        return

    @task
    def open_safety_shutter(self):
        return

    def safety_shutter_opened(self):
        return

    @task
    def close_safety_shutter(self):
        return

    @task
    def prepare_intensity_monitors(self):
        return

    def prepare_acquisition(
        self, take_dark, start, osc_range, exptime, npass, number_of_images, comment=""
    ):
        return

    def set_detector_filenames(
        self, frame_number, start, filename, jpeg_full_path, jpeg_thumbnail_full_path
    ):
        return

    def prepare_oscillation(self, start, osc_range, exptime, npass):
        return (start, start + osc_range)

    def do_oscillation(self, start, end, exptime, shutterless, npass, first_frame):
        gevent.sleep(exptime)

    def start_acquisition(self, exptime, npass, first_frame):
        return

    def write_image(self, last_frame):
        self.actual_frame_num += 1
        return

    def last_image_saved(self):
        return self.actual_frame_num

    def stop_acquisition(self):
        return

    def reset_detector(self):
        return

    def prepare_input_files(
        self, files_directory, prefix, run_number, process_directory
    ):
        self.actual_frame_num = 0
        i = 1
        while True:
            xds_input_file_dirname = "xds_%s_run%s_%d" % (prefix, run_number, i)
            xds_directory = os.path.join(process_directory, xds_input_file_dirname)

            if not os.path.exists(xds_directory):
                break

            i += 1

        mosflm_input_file_dirname = "mosflm_%s_run%s_%d" % (prefix, run_number, i)
        mosflm_directory = os.path.join(process_directory, mosflm_input_file_dirname)

        hkl2000_dirname = "hkl2000_%s_run%s_%d" % (prefix, run_number, i)
        hkl2000_directory = os.path.join(process_directory, hkl2000_dirname)

        self.raw_data_input_file_dir = os.path.join(
            files_directory, "process", xds_input_file_dirname
        )
        self.mosflm_raw_data_input_file_dir = os.path.join(
            files_directory, "process", mosflm_input_file_dirname
        )
        self.raw_hkl2000_dir = os.path.join(files_directory, "process", hkl2000_dirname)

        return xds_directory, mosflm_directory, hkl2000_directory

    @task
    def write_input_files(self, collection_id):
        return

    def get_wavelength(self):
        return

    def get_undulators_gaps(self):
        return []

    def get_resolution_at_corner(self):
        return

    def get_beam_size(self):
        return None, None

    def get_slit_gaps(self):
        return None, None

    def get_beam_shape(self):
        return

    def get_machine_current(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.get_current()
        else:
            return 0

    def get_machine_message(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.get_message()
        else:
            return ""

    def get_machine_fill_mode(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.get_fill_mode()
        else:
            """"""

    def get_cryo_temperature(self):
        if self.bl_control.cryo_stream is not None:
            return self.bl_control.cryo_stream.getTemperature()

    def get_current_energy(self):
        return

    def get_beam_centre(self):
        return None, None

    def get_beamline_configuration(self, *args):
        return self.bl_config._asdict()

    def is_connected(self):
        return True

    def is_ready(self):
        return True

    def sample_changer_HO(self):
        return self.bl_control.sample_changer

    def diffractometer(self):
        return self.bl_control.diffractometer

    def sanity_check(self, collect_params):
        return

    def set_brick(self, brick):
        return

    def directory_prefix(self):
        return self.bl_config.directory_prefix

    def store_image_in_lims(self, frame, first_frame, last_frame):
        return True

    def get_oscillation(self, oscillation_id):
        return self.oscillations_history[oscillation_id - 1]

    def sample_accept_centring(self, accepted, centring_status):
        self.sample_centring_done(accepted, centring_status)

    def set_centring_status(self, centring_status):
        self._centring_status = centring_status

    def get_oscillations(self, session_id):
        return []

    def set_helical(self, helical_on):
        return

    def set_helical_pos(self, helical_oscil_pos):
        return

    def get_archive_directory(self, directory):
        archive_dir = os.path.join(directory, "archive")
        return archive_dir

    @task
    def generate_image_jpeg(self, filename, jpeg_path, jpeg_thumbnail_path):
        pass



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
ALBAMiniDiff

[Description]
Specific HwObj for M2D2 diffractometer @ ALBA

[Channels]
- N/A

[Commands]
- N/A

[Emitted signals]
- pixelsPerMmChanged
- phiMotorMoved
- stateChanged
- zoomMotorPredefinedPositionChanged


[Functions]
- None

[Included Hardware Objects]
- None
"""

import time

import gevent

from mxcubecore.HardwareObjects.GenericDiffractometer import (
    DiffractometerState,
    GenericDiffractometer,
)
from mxcubecore.model import queue_model_objects

__author__ = "Jordi Andreu"
__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__maintainer__ = "Jordi Andreu"
__email__ = "jandreu[at]cells.es"
__status__ = "Draft"


class ALBAMiniDiff(GenericDiffractometer):
    """
    Specific diffractometer HwObj for XALOC beamline.
    """

    def __init__(self, *args):
        GenericDiffractometer.__init__(self, *args)
        self.centring_hwobj = None
        self.super_hwobj = None

    def init(self):
        self.calibration = self.get_object_by_role("calibration")

        self.centring_hwobj = self.get_object_by_role("centring")
        self.super_hwobj = self.get_object_by_role("beamline-supervisor")

        if self.centring_hwobj is None:
            self.log.debug("ALBAMinidiff: Centring math is not defined")

        if self.super_hwobj is not None:
            self.connect(
                self.super_hwobj, "stateChanged", self.supervisor_state_changed
            )
            self.connect(
                self.super_hwobj, "phaseChanged", self.supervisor_phase_changed
            )

        self.state_channel = self.get_channel_object("State")
        self.connect(self.state_channel, "update", self.state_changed)
        # This is not used
        self.cmd_start_auto_focus = self.get_command_object("startAutoFocus")

        self.phi_motor_hwobj = self.get_object_by_role("phi")
        self.phiz_motor_hwobj = self.get_object_by_role("phiz")
        self.phiy_motor_hwobj = self.get_object_by_role("phiy")
        self.zoom_motor_hwobj = self.get_object_by_role("zoom")
        self.focus_motor_hwobj = self.get_object_by_role("focus")
        self.sample_x_motor_hwobj = self.get_object_by_role("sampx")
        self.sample_y_motor_hwobj = self.get_object_by_role("sampy")

        if self.phi_motor_hwobj is not None:
            self.connect(
                self.phi_motor_hwobj, "stateChanged", self.phi_motor_state_changed
            )
            self.connect(self.phi_motor_hwobj, "valueChanged", self.phi_motor_moved)
        else:
            self.log.error("ALBAMiniDiff: Phi motor is not defined")

        if self.phiz_motor_hwobj is not None:
            self.connect(
                self.phiz_motor_hwobj, "stateChanged", self.phiz_motor_state_changed
            )
            self.connect(self.phiz_motor_hwobj, "valueChanged", self.phiz_motor_moved)
        else:
            self.log.error("ALBAMiniDiff: Phiz motor is not defined")

        if self.phiy_motor_hwobj is not None:
            self.connect(
                self.phiy_motor_hwobj, "stateChanged", self.phiy_motor_state_changed
            )
            self.connect(self.phiy_motor_hwobj, "valueChanged", self.phiy_motor_moved)
        else:
            self.log.error("ALBAMiniDiff: Phiy motor is not defined")

        if self.zoom_motor_hwobj is not None:
            self.connect(
                self.zoom_motor_hwobj, "valueChanged", self.zoom_position_changed
            )
            self.connect(
                self.zoom_motor_hwobj,
                "predefinedPositionChanged",
                self.zoom_motor_predefined_position_changed,
            )
            self.connect(
                self.zoom_motor_hwobj, "stateChanged", self.zoom_motor_state_changed
            )
        else:
            self.log.error("ALBAMiniDiff: Zoom motor is not defined")

        if self.sample_x_motor_hwobj is not None:
            self.connect(
                self.sample_x_motor_hwobj,
                "stateChanged",
                self.sampleX_motor_state_changed,
            )
            self.connect(
                self.sample_x_motor_hwobj, "valueChanged", self.sampleX_motor_moved
            )
        else:
            self.log.error("ALBAMiniDiff: Sampx motor is not defined")

        if self.sample_y_motor_hwobj is not None:
            self.connect(
                self.sample_y_motor_hwobj,
                "stateChanged",
                self.sampleY_motor_state_changed,
            )
            self.connect(
                self.sample_y_motor_hwobj, "valueChanged", self.sampleY_motor_moved
            )
        else:
            self.log.error("ALBAMiniDiff: Sampx motor is not defined")

        if self.focus_motor_hwobj is not None:
            self.connect(self.focus_motor_hwobj, "valueChanged", self.focus_motor_moved)

        GenericDiffractometer.init(self)

        queue_model_objects.CentredPosition.set_diffractometer_motor_names(
            "phi", "phiy", "phiz", "sampx", "sampy", "kappa"
        )

    def state_changed(self, state):
        """
        Overrides method to map Tango ON state to Difractaometer State Ready.

        @state: Tango state
        """

        if str(state) == "ON":
            state = DiffractometerState.tostring(DiffractometerState.Ready)

        if state != self.current_state:
            self.log.debug(
                "ALBAMinidiff: State changed %s (was: %s)"
                % (str(state), self.current_state)
            )
            self.current_state = state
            self.emit("minidiffStateChanged", (self.current_state))

    def getCalibrationData(self, offset=None):
        """
        Get pixel size for OAV system

        @offset: Unused
        @return: 2-tuple float
        """
        calibx, caliby = self.calibration.getCalibration()
        return 1000.0 / caliby, 1000.0 / caliby
        # return 1000./self.md2.CoaxCamScaleX, 1000./self.md2.CoaxCamScaleY

    def get_pixels_per_mm(self):
        """
        Returns the pixel/mm for x and y. Overrides GenericDiffractometer method.
        """
        px_x, px_y = self.getCalibrationData()
        return (px_x, px_y)

    def update_pixels_per_mm(self, *args):
        """
        Emit signal with current pixel/mm values.
        """
        self.pixels_per_mm_x, self.pixels_per_mm_y = self.getCalibrationData()
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),))

    # TODO: looks quite bizarre.
    def motor_positions_to_screen(self, centred_positions_dict):
        """
        Convert motor positions contained in a dictionary to screen (canvas) positions.
        Overrides GenericDiffractometer method.

        @centered_positions_dict: dictionary to be converted
        @return: position
        """
        c = centred_positions_dict

        # if self.head_type == GenericDiffractometer.HEAD_TYPE_MINIKAPPA:
        # kappa = self.motor_hwobj_dict["kappa"]
        # phi = self.motor_hwobj_dict["kappa_phi"]

        xy = self.centring_hwobj.centringToScreen(c)
        if xy is None:
            return None

        x = xy["X"] * self.pixels_per_mm_x + self.zoom_centre["x"]

        y = xy["Y"] * self.pixels_per_mm_y + self.zoom_centre["y"]

        # self.log.debug("  motor_positions_to_screen ")
        # self.log.debug(" positions = %s " % str(centred_positions_dict))
        # self.log.debug(" x,y = %s, %s " % (x,y))

        return x, y

    # TODO: Must be implemented correctly.
    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """
        Returns a dictionary with motors name and positions centred.
        It is expected in start_move_to_beam and move_to_beam methods in
        GenericDIffractometer HwObj.

        @return: dict
        """
        return {"omega": [200, 200]}

    def getBeamInfo(self, update_beam_callback):
        """
        Update beam info (position and shape) and execute callback.

        @update_beam_callback: callback method passed as argument.
        """
        calibx, caliby = self.calibration.getCalibration()

        size_x = self.get_channel_object("beamInfoX").get_value() / 1000.0
        size_y = self.get_channel_object("beamInfoY").get_value() / 1000.0

        data = {"size_x": size_x, "size_y": size_y, "shape": "ellipse"}

        update_beam_callback(data)

    # TODO:Implement dynamically
    def use_sample_changer(self):
        """
        Overrides GenericDiffracometer method.
        """
        return True

    # TODO:Implement dynamically
    def in_plate_mode(self):
        """
        Overrides GenericDiffracometer method.
        """
        return False

    # We are using the sample_centring module. this is not used anymore
    def start_manual_centring(self, *args, **kwargs):
        """
        Start manual centring. Overrides GenericDiffracometer method.
        Prepares diffractometer for manual centring.
        """
        if self.prepare_centring():
            GenericDiffractometer.start_manual_centring(self, *args, **kwargs)
        else:
            self.log.info(" Failed to prepare diffractometer for centring")
            self.invalidate_centring()

    def start_auto_centring(self, *args, **kwargs):
        """
        Start manual centring. Overrides GenericDiffracometer method.
        Prepares diffractometer for manual centring.
        """
        if self.prepare_centring():
            GenericDiffractometer.start_auto_centring(self, *args, **kwargs)
        else:
            self.log.info(" Failed to prepare diffractometer for centring")
            self.invalidate_centring()

    def prepare_centring(self):
        """
        Prepare beamline for to sample_view phase.
        """
        if not self.is_sample_view_phase():
            self.log.info(" Not in sample view phase. Asking supervisor to go")
            success = self.go_sample_view()
            if not success:
                self.log.info("Cannot set SAMPLE VIEW phase")
                return False

        return True

    # def manual_centring(self):
    #    """
    #    We are using the sample_centring module. this is not used anymore
    #    """
    #    self.centring_hwobj.initCentringProcedure()

    #    # self.head_type = self.chan_head_type.get_value()
    #    # Say diffractometer to go to SampleView phase

    #    # go to sample_view phase
    #    if not self.is_sample_view_phase():
    #        self.log.info(" Not in sample view phase. Asking supervisor to go")
    #        success = self.go_sample_view()
    #        if not success:
    #            self.log.info("Cannot set SAMPLE VIEW phase")
    #            return False

    #    phi_init_position = self.phi_motor_hwobj.get_value()

    #    for click in range(3):
    #        self.user_clicked_event = gevent.event.AsyncResult()
    #        x, y = self.user_clicked_event.get()
    #        self.centring_hwobj.appendCentringDataPoint(
    #             {"X": (x - self.zoom_centre['x'])/ self.pixels_per_mm_x,
    #              "Y": (y - self.zoom_centre['y'])/ self.pixels_per_mm_y})

    #        if self.in_plate_mode():
    #            dynamic_limits = self.phi_motor_hwobj.get_dynamic_limits()
    #            if click == 0:
    #                self.phi_motor_hwobj.set_value(dynamic_limits[0])
    #            elif click == 1:
    #                self.phi_motor_hwobj.set_value(dynamic_limits[1])
    #        else:
    #            if click < 2:
    #                self.phi_motor_hwobj.set_value_relative(-90, timeout=None)

    #    #self.log.info(" Returning phi to initial position %s" % phi_init_position)
    #    #self.phi_motor_hwobj.set_value(phi_init_position, timeout=None)
    #
    #    return self.centring_hwobj.centeredPosition(return_by_name=False)

    def is_sample_view_phase(self):
        """
        Returns boolean by comparing the supervisor current phase and SAMPLE view phase.

        @return: boolean
        """
        return self.super_hwobj.get_current_phase().upper() == "SAMPLE"

    def go_sample_view(self):
        """
        Go to sample view phase.
        """
        self.super_hwobj.go_sample_view()

        while True:
            super_state = str(self.super_hwobj.get_state()).upper()
            if super_state != "MOVING":
                self.log.debug(
                    "ALBAMinidiff: go_sample_view done . super_state is %s"
                    % super_state
                )
                return True
            gevent.sleep(0.2)

    def supervisor_state_changed(self, state):
        """
        Emit stateChanged signal according to supervisor current state.
        """
        return
        self.current_state = state
        self.emit("stateChanged", (self.current_state,))

    # TODO: Review override current_state by current_phase
    def supervisor_phase_changed(self, phase):
        """
        Emit stateChanged signal according to supervisor current phase.
        """
        # self.current_state = phase
        self.emit("minidiffPhaseChanged", (phase,))

    def phi_motor_moved(self, pos):
        """
        Emit phiMotorMoved signal with position value.
        """
        self.current_motor_positions["phi"] = pos
        # self.emit_diffractometer_moved()
        self.emit("phiMotorMoved", pos)
        # self.emit('stateChanged', (self.current_motor_states["phi"], ))

    def phi_motor_state_changed(self, state):
        """
        Emit stateChanged signal with state value.
        """
        self.current_motor_states["phi"] = state
        self.emit("stateChanged", (state,))

    def phiz_motor_moved(self, pos):
        """ """
        self.current_motor_positions["phiz"] = pos
        # if time.time() - self.centring_time > 3.0:
        #    self.invalidate_centring()
        # self.emit_diffractometer_moved()

    def phiz_motor_state_changed(self, state):
        """
        Emit stateChanged signal with state value.
        """
        self.emit("stateChanged", (state,))

    def phiy_motor_state_changed(self, state):
        """
        Emit stateChanged signal with state value.
        """
        self.emit("stateChanged", (state,))

    def phiy_motor_moved(self, pos):
        """ """
        self.current_motor_positions["phiy"] = pos
        # if time.time() - self.centring_time > 3.0:
        #    self.invalidate_centring()
        # self.emit_diffractometer_moved()

    def zoom_position_changed(self, value):
        """
        Update positions after zoom changed.

        @value: zoom position.
        """
        self.update_pixels_per_mm()
        self.current_motor_positions["zoom"] = value
        self.refresh_omega_reference_position()

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        """
        Update pixel size and emit signal.
        """
        self.update_pixels_per_mm()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def zoom_motor_state_changed(self, state):
        """
        Emit signal for motor zoom changed

        @state: new state value to emit.
        """
        self.emit("stateChanged", (state,))

    def sampleX_motor_moved(self, pos):
        """ """
        self.current_motor_positions["sampx"] = pos
        # if time.time() - self.centring_time > 3.0:
        #    self.invalidate_centring()
        # self.emit_diffractometer_moved()

    def sampleX_motor_state_changed(self, state):
        """
        Emit stateChanged signal with state value.
        """
        self.current_motor_states["sampx"] = state
        self.emit("stateChanged", (state,))

    def sampleY_motor_moved(self, pos):
        """ """
        self.current_motor_positions["sampy"] = pos
        # if time.time() - self.centring_time > 3.0:
        #    self.invalidate_centring()
        # self.emit_diffractometer_moved()

    def sampleY_motor_state_changed(self, state):
        """
        Emit stateChanged signal with state value.
        """
        self.current_motor_states["sampy"] = state
        self.emit("stateChanged", (state,))

    def focus_motor_moved(self, pos):
        """ """
        self.current_motor_positions["focus"] = pos

    # TODO:  The command is not configured in the xml. Unused
    def start_auto_focus(self):
        self.cmd_start_auto_focus()

    def move_omega(self, pos, velocity=None):
        """
        Move omega to absolute position.

        @pos: target position
        """
        # turn it on
        if velocity is not None:
            self.phi_motor_hwobj.set_velocity(velocity)
        self.phi_motor_hwobj.set_value(pos)
        time.sleep(0.2)
        # it should wait here

    def move_omega_relative(self, relpos):
        """
        Move omega to relative position.

        @relpos: target relative position
        """
        self.wait_device_ready()
        self.phi_motor_hwobj.set_value_relative(relpos, timeout=None)

    # TODO: define phases as enum members.
    def set_phase(self, phase):
        """
        General function to set phase by using supervisor commands.
        """
        if phase == "Transfer":
            self.super_hwobj.go_transfer()
        elif phase == "Collect":
            self.super_hwobj.go_collect()
        elif phase == "BeamView":
            self.super_hwobj.go_beam_view()
        elif phase == "Centring":
            self.super_hwobj.go_sample_view()
        else:
            self.log.warning(
                "Diffractometer set_phase asked for un-handled phase: %s" % phase
            )


def test_hwo(hwo):
    print(hwo.get_phase_list())



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract hardware object class for characterisation."""

import abc

import gevent.event

from mxcubecore.BaseHardwareObjects import HardwareObject


class AbstractCharacterisation(HardwareObject):
    """Abstract hardware object class for characterisation."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name) -> None:
        super().__init__(name)
        self.processing_done_event = gevent.event.Event()

    @abc.abstractmethod
    def characterise(self, _input):
        """Start the characterosation.

        Args:
            input (object): Characterisation input object.

        Returns:
            (str) The Characterisation result.
        """

    @abc.abstractmethod
    def get_html_report(self, output):
        """Get the path to the characterisation html report, generated by the
        characterisation software.

        Args:
            output (object): Characterisation output object

        Returns:
            (str) The path to the html result report.
        """

    @abc.abstractmethod
    def input_from_params(self, ref_parameters, char_params, path_str):
        """
        Args:
            ref_parameters:
                A named tuple or object with following fields:

                'id',
                'prefix',
                'run_number',
                'template',
                'first_image',
                'num_images',
                'osc_start',
                'osc_range',
                'overlap',
                'exp_time',
                'num_passes',
                'comments',
                'path',
                'centred_positions',
                'energy',
                'resolution',
                'transmission',
                'shutterless',
                'inverse_beam',
                'screening_id'

            char_params:
                A named tuple or object with following fields:

                # Optimisation parameters
                'aimed_resolution'
                'aimed_multiplicity'
                'aimed_i_sigma'
                'aimed_completness'
                'strategy_complexity'
                'induce_burn'
                'use_permitted_rotation'
                'permitted_phi_start'
                'permitted_phi_end'

                # Crystal
                'max_crystal_vdim'
                'min_crystal_vdim'
                'max_crystal_vphi'
                'min_crystal_vphi'
                'space_group'

                # Characterisation type
                'use_min_dose'
                'use_min_time'
                'min_dose'
                'min_time'
                'account_rad_damage'
                'not_use_low_res'
                'auto_res'
                'opt_sad'
                'sad_res'
                'determine_rad_params'

                # Radiation damage model
                'rad_suscept'
                'beta'
                'sigma'

            path_str: Template string representing path to each image.

        Returns:
            Input for characterisation software.
        """

    @abc.abstractmethod
    def dc_from_output(self, output):
        """Create a data collection from characterisation result

        Args:
            output (object): Characterisation result object

        Returns:
            (queue_model_objects.DataCollection)
        """

    @abc.abstractmethod
    def get_default_characterisation_parameters(self, default_input_file):
        """
        Args:
            default_input_file (str): Path to file containing default input.

        Returns:
            (queue_model_objects.CharacterisationsParameters): object with
            default parameters.
        """

    def prepare_input(self, _input):
        """Called by characterise before characterisation starts.

        Can be used to manipulate edna_input object before characterisation starts.
        Example: to set a site specific output directory

        Args:
           _input (object): Characterisation input object.
        """

    def is_running(self) -> bool:
        """
        Returns:
            ``True`` if process is running otherwise ``False``.
        """
        return not self.processing_done_event.is_set()



from mxcubecore.model import queue_model_objects
from mxcubecore.queue_entry import *


class PX2DataCollectionQueueEntry(DataCollectionQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        DataCollectionQueueEntry.__init__(self, view, data_model, view_set_queue_entry)

        self.collect_task = None
        self.centring_task = None
        self.session = None

    def collect_dc(self, dc, list_item):
        log = logging.getLogger("user_level_log")

        log.info(
            "queue_entry. Start data collection on object %s"
            % str(HWR.beamline.collect)
        )

        if HWR.beamline.collect:
            acq_1 = dc.acquisitions[0]
            cpos = acq_1.acquisition_parameters.centred_position
            # sample = self.get_data_model().get_parent().get_parent()
            sample = self.get_data_model().get_sample_node()

            try:
                if dc.experiment_type is EXPERIMENT_TYPE.HELICAL:
                    acq_1, acq_2 = (dc.acquisitions[0], dc.acquisitions[1])
                    # HWR.beamline.collect.get_channel_object("helical").set_value(1)

                    start_cpos = acq_1.acquisition_parameters.centred_position
                    end_cpos = acq_2.acquisition_parameters.centred_position

                    helical_oscil_pos = {
                        "1": start_cpos.as_dict(),
                        "2": end_cpos.as_dict(),
                    }
                    # HWR.beamline.collect.get_channel_object('helical_pos').set_value(helical_oscil_pos)
                    HWR.beamline.collect.set_helical(True, helical_oscil_pos)

                    msg = "Helical data collection, moving to start position"
                    log.info(msg)
                    log.info("Moving sample to given position ...")
                    list_item.setText(1, "Moving sample")
                else:
                    # HWR.beamline.collect.get_channel_object("helical").set_value(0)
                    HWR.beamline.collect.set_helical(False)

                empty_cpos = queue_model_objects.CentredPosition()

                if cpos != empty_cpos:
                    log.info("Moving sample to given position ...")
                    list_item.setText(1, "Moving sample")
                    HWR.beamline.sample_view.select_shape_with_cpos(cpos)
                    self.centring_task = (
                        HWR.beamline.diffractometer.moveToCentredPosition(
                            cpos, wait=False
                        )
                    )
                    self.centring_task.get()
                else:
                    pos_dict = HWR.beamline.diffractometer.get_positions()
                    cpos = queue_model_objects.CentredPosition(pos_dict)
                    snapshot = HWR.beamline.sample_view.get_snapshot([])
                    acq_1.acquisition_parameters.centred_position = cpos
                    acq_1.acquisition_parameters.centred_position.snapshot_image = (
                        snapshot
                    )

                param_list = queue_model_objects.to_collect_dict(
                    dc,
                    sample,
                    cpos if cpos != empty_cpos else None,
                )
                self.collect_task = HWR.beamline.collect.collect(
                    COLLECTION_ORIGIN_STR.MXCUBE, param_list
                )
                self.collect_task.get()

                if "collection_id" in param_list[0]:
                    dc.id = param_list[0]["collection_id"]

                dc.acquisitions[0].path_template.xds_dir = param_list[0]["xds_dir"]

            except gevent.GreenletExit:
                # log.warning("Collection stopped by user.")
                list_item.setText(1, "Stopped")
                raise QueueAbortedException("queue stopped by user", self)
            except Exception as ex:
                self.log.exception("")
                raise QueueExecutionException(ex.message, self)
        else:
            log.error(
                "Could not call the data collection routine,"
                + " check the beamline configuration"
            )
            list_item.setText(1, "Failed")
            msg = (
                "Could not call the data collection"
                + " routine, check the beamline configuration"
            )
            raise QueueExecutionException(msg, self)


class PX2EnergyScanQueueEntry(EnergyScanQueueEntry):
    def __init__(self, view=None, data_model=None):
        EnergyScanQueueEntry.__init__(self, view, data_model)
        self.energy_scan_task = None
        self._failed = False

    def energy_scan_finished(self, scan_info):
        energy_scan = self.get_data_model()
        scan_file_path = os.path.join(
            energy_scan.path_template.directory, energy_scan.path_template.get_prefix()
        )
        logging.info(
            "HWR.beamline.energy_scan %s type %s"
            % (HWR.beamline.energy_scan, type(HWR.beamline.energy_scan))
        )
        logging.info("energy_scan %s type %s" % (energy_scan, type(energy_scan)))
        scan_file_archive_path = os.path.join(
            energy_scan.path_template.get_archive_directory(),
            energy_scan.path_template.get_prefix(),
        )
        logging.info(
            "energy_scan.element_symbol %s, energy_scan.edge %s, scan_file_archive_path %s, scan_file_path %s"
            % (
                energy_scan.element_symbol,
                energy_scan.edge,
                scan_file_archive_path,
                scan_file_path,
            )
        )
        egy_result = HWR.beamline.energy_scan.do_chooch(
            energy_scan.element_symbol,
            energy_scan.edge,
            scan_file_archive_path,
            scan_file_path,
        )

        if egy_result is None:
            logging.info("energy_scan. failed. ")
            return None

        (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        ) = egy_result

        # scan_info = HWR.beamline.energy_scan.scanInfo

        # This does not always apply, update model so
        # that its possible to access the sample directly from
        # the EnergyScan object.
        sample = self.get_view().parent().parent().get_model()
        sample.crystals[0].energy_scan_result.peak = pk
        sample.crystals[0].energy_scan_result.inflection = ip
        sample.crystals[0].energy_scan_result.first_remote = rm
        sample.crystals[0].second_remote = None

        energy_scan.result = sample.crystals[0].energy_scan_result

        logging.getLogger("user_level_log").info(
            "Energy scan, result: peak: %.4f, inflection: %.4f"
            % (
                sample.crystals[0].energy_scan_result.peak,
                sample.crystals[0].energy_scan_result.inflection,
            )
        )

        self.get_view().setText(1, "Done")



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Micro Diffractometer implementation of the AbstractDiffractometer class."""

from gevent import Timeout, sleep

from mxcubecore import HardwareRepository as HWR
from mxcubecore.Command.Exporter import Exporter
from mxcubecore.Command.exporter.ExporterStates import ExporterStates
from mxcubecore.HardwareObjects.abstract.AbstractDiffractometer import (
    AbstractDiffractometer,
    DiffractometerConstraint,
    DiffractometerHead,
    DiffractometerPhase,
)

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicroDiffractometer(AbstractDiffractometer):
    """Microdiff with Exporter implementation of AbstractDiffractometer"""

    def __init__(self, name):
        super().__init__(name)
        self._exporter = None

    def init(self):
        """Initialise the device"""
        super().init()
        exporter_address = self.get_property("exporter_address")
        _host, _port = exporter_address.split(":")
        self._exporter = Exporter(_host, int(_port))
        self.head_type = self._get_head_type

        # add the custom commands
        for nam, cmd in self.get_property("commands").items():
            _cmd = {
                "type": "exporter",
                "exporter_address": exporter_address,
                "name": nam,
            }
            setattr(self, nam, self.add_command(_cmd, cmd))

        # add the custom channels
        for nam, attr in self.get_property("channels").items():
            _attr = {
                "type": "exporter",
                "exporter_address": exporter_address,
                "name": nam,
            }
            setattr(self, nam, self.add_channel(_attr, attr))

        self.update_state()

        # we must have global_state and phase_channel channels configured
        try:
            self.phase_channel.connect_signal("update", self.update_phase)
            self.global_state.connect_signal("update", self._update_state)
        except AttributeError:
            self.log.exception("global_state and phase_channel not configured!")

    def abort(self):
        """Immediately terminate action."""
        self._exporter.execute("abort")

    @property
    def _get_hwstate(self) -> str:
        """Get the hardware state, reported by the MD2 application.

        Returns:
            The state.
        """
        try:
            return self._exporter.read_property("HardwareState")
        except AttributeError:
            return "Ready"

    @property
    def _get_swstate(self) -> str:
        """Get the software state, reported by the MD2 application.

        Returns:
            The state.
        """
        return self.global_state.get_value()

    def _update_state(self, value):
        if isinstance(value, str):
            self.update_state(self.get_state())
        else:
            self.update_state(value)

    def get_state(self):
        """Get the diffractometer general state.

        Returns:
            (enum 'HardwareObjectState'): state
        """
        try:
            return ExporterStates[self._get_swstate.upper()].value
        except ValueError:
            return self.STATES.UNKNOWN

    @property
    def _ready(self) -> bool:
        """Get the "Ready" state - software and hardware.

        Returns:
            True if both "Ready", False otherwise.
        """
        return self._get_swstate == "Ready" and self._get_hwstate == "Ready"

    def wait_status_ready(self, timeout: float | None = None):
        """Wait timeout seconds until status is ready.

        Args:
            Timeout [s]. None means infinite timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for status ready")):
            while not self._ready:
                sleep(0.5)

    def set_value_motors(
        self,
        motors_positions_dict: dict[str, float],
        simultaneous: bool = True,
        timeout: float | None = None,
    ):
        """Move specified motors to the requested positions.

        Args:
            motors_positions_dict (dict): Dictionary {motor_role: target_value}.
            simultaneous: Move the motors simultaneously (True - default) or not.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            TimeoutError: Timeout
            KeyError: The name does not correspond to an existing motor
        """
        if not simultaneous:
            super().set_value_motors(motors_positions_dict, simultaneous, timeout)
        else:
            # prepare the command
            cmd = ""
            for role, pos in motors_positions_dict.items():
                name = self.motors_hwobj_dict[role].actuator_name
                cmd += f"{name}={pos:0.3f};"
            self._exporter.execute("startSimultaneousMoveMotors", (cmd,))
            if timeout != 0:
                self.wait_status_ready(timeout)
            self.update_state()

    def get_value_motors(self, motors_list: [list | None] = None) -> dict[str, float]:
        """Get the positions of diffractometer motors. If the motors_list
           is empty, return the positions of all the availble motors.

        Args:
            motors_list: List of motor roles.

        Returns:
            Dictionary {role: position}.
        """
        motors_positions_dict = super().get_value_motors(motors_list)
        if not self.in_kappa_mode:
            motors_positions_dict.update({"kappa": None, "kappa_phi": None})
        return motors_positions_dict

    def get_motors(self):
        """Get the dictionary of all the motors which can be used.

        Returns:
            Ddctionary {role: hardware_object}.
        """

        def find_elem(ddict, val):
            """Find dictionary elemnt from motor actuator_name"""
            for role, hwobj in ddict.items():
                if hwobj.actuator_name == val:
                    return {role: hwobj}
            return {}

        motors = self._exporter.read_property("MotorStates")
        motors_dict = {}
        for mot in motors:
            mot_stat = mot.split("=")
            if mot_stat[1] not in ("Disable", "Unknown"):
                elem = find_elem(self.motors_hwobj_dict, mot_stat[0])
                motors_dict.update(elem)

        return motors_dict

    @property
    def _get_head_type(self) -> DiffractometerHead:
        """Get the head type."""
        try:
            self.head_type = DiffractometerHead(
                self._exporter.read_property("HeadType")
            )
        except ValueError:
            self.head_type = DiffractometerHead.UNKNOWN
        return self.head_type

    def _set_phase(self, value: DiffractometerPhase):
        """Specific implementation to set the diffractometer to selected phase.

        Args:
            value: requested phase.
        """
        _use_custom = self.get_property("use_custom_phase_script") or False
        current_phase = self.get_phase()

        if value != current_phase:
            msg = f"Current phase is {current_phase} and moving to {value}"
            self.log.info(msg)

        if _use_custom and not self.in_plate_mode:
            script = "ChangePhase_" + value.value.lower()
            msg = f"Changing phase to {value.value}, using pmac script"
            self.log.info(msg)
            self.run_custom_script(script)
        else:
            self._exporter.execute("startSetPhase", (value.value,))
        self.wait_status_ready(timeout=200)

    def get_phase(self) -> DiffractometerPhase:
        """Get the current phase."""
        value = self.phase_channel.get_value()
        try:
            self.current_phase = DiffractometerPhase(value)
        except ValueError:
            self.current_phase = DiffractometerPhase.UNKNOWN
        return self.current_phase

    def _set_constraint(self, value: DiffractometerConstraint):
        """Specific implementation to set the diffractometer to selected constraint
        Args:
            value: requested constraint.
        """
        self._exporter.execute("startSetMode", (value.value,))

    def get_constraint(self) -> DiffractometerConstraint:
        """Get the diffrractometer constraint type."""
        value = self._exporter.read_property("CurrentMode")
        try:
            self.current_constraint = DiffractometerConstraint(value)
        except ValueError:
            self.current_constraint = DiffractometerConstraint.UNKNOWN
        return self.current_constraint

    def check_scan_limits(self, start: float, end: float, exptime: float) -> bool:
        """Check if the scan parameters are within the limits

        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total) [s].

        Returns:
            True (parameters within the limits), False otherwise.
        """
        if self.in_plate_mode:
            scan_speed = abs(end - start) / exptime
            llim, hlim = map(
                float,
                self._exporter.execute("getOmegaMotorDynamicScanLimits", (scan_speed,)),
            )
            if start < llim:
                msg = f"Scan start below the allowed value {llim}"
                raise ValueError(msg)
            if end > hlim:
                msg = f"Scan end above the allowed value {hlim}"
                raise ValueError(msg)
        return True

    def do_oscillation_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        number_of_images: int = 1,
        timeout: float | None = None,
    ):
        """Do an oscillation scan on omega.

        Args:
            start: omega start position.
            end: omega end position.
            exptime: scan exposure time (total).
            number_of_images: Used if need to set number of frames.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
            ValueError: Scan parameters not within limits (if relevant).
        """
        # check the scan limits
        self.check_scan_limits(start, end, exptime)
        # set the number of frames
        if not self.get_property("md_set_number_of_frames"):
            number_of_images = 1
        self._exporter.write_property("ScanNumberOfFrames", number_of_images)
        scan_params = f"1\t{start:0.3f}\t{(end - start):0.3f}\t{exptime:0.3f}\t1"
        self._exporter.execute("startScanEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_line_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        number_of_images: int,
        motors_pos: dict[str, dict],
        timeout: float | None = None,
    ):
        """Do helical (line) scan on omega.
        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total).
            number_of_images: Used only if more tahn one frame needed.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).
            motors_pos: {"1": [centred position], "2": [centred position]}

        Raises:
            RuntimeError: Timeout waiting for status ready.
            ValueError: Scan parameters not within limits (if relevant).
        """
        # check the scan limits
        self.check_scan_limits(start, end, exptime)
        if not self.get_property("md_set_number_of_frames"):
            number_of_images = 1

        self._exporter.write_property("ScanNumberOfFrames", number_of_images)

        scan_params = f"{start:0.3f}\t{(end - start):0.3f}\t{exptime:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            scan_params += f"{motors_pos['1'][name]:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            scan_params += f"{motors_pos['2'][name]:0.3f}\t"

        self._exporter.execute("startScan4DEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_mesh_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        dead_time: float,
        nb_lines: int,
        nb_frames_total: int,
        grid_centre: list[tuple[str, float]],
        mesh_range: dict,
        timeout: float | None = None,
    ):
        """Do a mesh scan.

        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total).
            dead_time: Dead time between the pulses. Detector dependant.
            nb_lines: Total number of lines.
            nb_frames_total: Total number of frames.
            grid_centre: List of tuples (motor_role, position).
                         representing the centre of the mesh grid.
            mesh_range: Horizontal and vertical range.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """

        # enable gate pulses
        self._exporter.write_property("DetectorGatePulseEnabled", value=True)

        # dead_time depends on the detector. We transform it to us
        dead_time = dead_time or HWR.beamline.detector.get_deadtime() * 1000
        self._exporter.write_property("DetectorGatePulseReadoutTime", dead_time)

        grid_centre = grid_centre.as_dict()
        self.set_value_motors(grid_centre, simultaneous=True, timeout=timeout)

        scan_params = f"{(end - start):0.3f}\t"
        scan_params += f"{-mesh_range['horizontal_range']:0.3f}\t"
        scan_params += f"{mesh_range['vertical_range']:0.3f}\t"
        scan_params += f"{start:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            for key, val in grid_centre.items():
                if name == key:
                    scan_params += f"{float(val):0.3f}\t"
        scan_params += f"{nb_lines}\t"
        scan_params += f"{nb_frames_total / nb_lines}\t"
        scan_params += f"{exptime / nb_lines}\t"
        scan_params += "True\tTrue\tTrue\t"
        self._exporter.execute("startRasterScanEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_still_scan(
        self,
        pulse_duration: float,
        pulse_period: float,
        nb_pulse: int,
        timeout: [None | float] = None,
    ):
        """Do a zero oscillation acquisition.

        Args:
            pulse_duration: Duration of the pulse sent to the detector.
            pulse_period: The period of the pulse sent to the detector.
            nb_pulse: Number of pulses to be sent.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """
        scan_params = f"{pulse_duration:0.6f}\t{pulse_period:0.6f}\t{nb_pulse}"
        self._exporter.execute("startStillScan", (scan_params,))
        self.wait_status_ready(timeout)

    def do_characterisation_scan(
        self,
        start: float,
        scan_range: float,
        nb_frames: int,
        exptime: float,
        nb_scans: int,
        angle: float,
        timeout: float | None = None,
    ):
        """Do fast characterisation.
        Args:
            start: Position of omega for the first scan [deg].
            scan_range: range for each scan [deg].
            nb_frames: Frame numbers for each scan.
            exptime: Total exposure time for each scan [s].
            nb_scans: How many times a scan to be repeated.
            angle: The angle between each scan [deg]. This number,
                   added to the last position of each scan and will
                   be the start position of the following scan.
            timeout (float): optional - timeout [s],
                             if timeout = 0: return at once and do not wait,
                             if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """

        if self.in_plate_mode:
            # to see if needed when plates
            return

        scan_params = f"{nb_frames}\t{start:0.3f}\t{scan_range:0.3f}\t"
        scan_params += f"{exptime:0.3f}\t{nb_scans}\t{angle:0.3f}"
        self._exporter.execute("startCharacterisationScanEx", (scan_params,))
        if timeout:
            # min timeout is 20 min
            timeout = max(timeout, 20 * 60)

        self.wait_status_ready(timeout)

    def get_pixels_per_mm(self) -> tuple[int, int]:
        """Get the pixel/mm values.

        Returns:
            (x ,y) [pixel/mm]
        """
        x_calib = self._exporter.read_property("CoaxCamScaleX")
        y_calib = self._exporter.read_property("CoaxCamScaleY")
        return 1.0 / x_calib, 1.0 / y_calib

    def get_beam_position(self) -> tuple:
        """Get the beam position defined in MD"""
        return (
            self.beam_position_horizontal.get_value(),
            self.beam_position_vertical.get_value(),
        )

    def run_custom_script(self, script_cmd: str, timeout: float | None = None):
        """Run custom script."""
        self.run_script(script_cmd)
        self.wait_status_ready(timeout)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


import abc

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2010- 2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractAuthenticator(HardwareObject):
    __metaclass__ = abc.ABCMeta

    def init(self) -> None:
        super().init()

    @abc.abstractmethod
    def authenticate(self, username: str, password: str) -> bool:
        """
        Authenticate with credentials username, password

        Args:
            username: username
            password: password

        Returns:
            True on success otherwise false
        """
        pass

    @abc.abstractmethod
    def invalidate(username: str) -> None:
        """
        de-authetnicate user with <username>

        Args:
            username: username
        """



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


"""Abstract hardware object for the aperture."""

from __future__ import annotations

import ast
from warnings import warn

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractAperture(HardwareObject):
    """Abstract hardware object for the aperture.

    Emits:
        diameterIndexChanged (int, float):
            Two-item tuple: current index and diameter in millimeters.
        valueChanged (str):
            Current position name.
    """

    def __init__(self, name: str) -> None:
        warn(
            "AbstractAperture is deprecated. Use AbstractNState instead",
            DeprecationWarning,
        )
        HardwareObject.__init__(self, name)

        self._current_position_name = None
        self._current_diameter_index = None
        self._diameter_size_list = []
        self._position_list = []

    def init(self):
        self._diameter_size_list = self.get_property("diameter_size_list", [])
        if isinstance(self._diameter_size_list, str):
            self._diameter_size_list = ast.literal_eval(self._diameter_size_list)
        if not self._diameter_size_list:
            self.log.warning("diameter_size_list is empty")

        self._position_list = self.get_property("position_list", [])
        if isinstance(self._position_list, str):
            self._position_list = ast.literal_eval(self._position_list)
        if not self._position_list:
            self.log.warning("position_list is empty")

    def get_diameter_size_list(self) -> list[float]:
        """Get list of diameter sizes.

        Returns:
            List of diameter sizes in microns.
        """
        return self._diameter_size_list

    def get_position_list(self) -> list[str]:
        """Get list of positions.

        Returns:
            Position names as a list of strings.
        """
        return self._position_list

    def get_diameter_index(self) -> int:
        """Get current diameter index.

        Returns:
            Current diameter index.
        """
        return self._current_diameter_index

    def set_diameter_index(self, diameter_index: int) -> None:
        """Set active diameter index.

        Args:
            diameter_index: Selected diameter index.
        """
        if diameter_index < len(self._diameter_size_list):
            self._current_diameter_index = diameter_index
            self.emit(
                "diameterIndexChanged",
                self._current_diameter_index,
                self._diameter_size_list[self._current_diameter_index] / 1000.0,
            )
        else:
            self.log.warning(
                "Aperture: Diameter index %d is not valid" % diameter_index
            )

    def get_diameter_size(self) -> float:
        """Get diameter size.

        Returns:
            Current diameter size in millimeters.
        """
        return self._diameter_size_list[self._current_diameter_index]

    def set_diameter_size(self, diameter_size: int) -> None:
        """Set diameter size.

        Args:
            diameter_size: selected diameter index
        """
        if diameter_size in self._diameter_size_list:
            self.set_diameter_index(self._diameter_size_list.index(diameter_size))
        else:
            self.log.warning("Aperture: Selected diameter is not in the diameter list")

    def get_position_name(self) -> str:
        """Get current position name.

        Returns:
            Current position name.
        """
        return self._current_position_name

    def set_position(self, position_index: int) -> None:
        warn(
            "set_position is deprecated. Use set_position_index(position_index) instead",
            DeprecationWarning,
        )
        self.set_position_index(position_index)

    def set_position_name(self, position_name: str) -> None:
        """Set aperture position based on a position name.

        Args:
            position_name: Selected position name.
        """
        if position_name in self._position_list:
            self._current_position_name = position_name
            self.emit("valueChanged", self._current_position_name)
        else:
            self.log.warning(
                "Aperture: Position %s is not in the position list" % position_name
            )

    def set_position_index(self, position_index: int) -> None:
        """Set aperture position based on a position index.

        Args:
            position_index: Selected position index.
        """
        if position_index < len(self._position_list):
            self._current_position_name = self._position_list[position_index]
            self.emit("valueChanged", self._current_position_name)
        else:
            self.log.warning("Aperture: Selected position is not in the position list")

    def set_in(self):
        """Set aperture in the beam."""

    def set_out(self):
        """Remove aperture from the beam."""

    def is_out(self) -> bool:
        """
        Returns:
            ``True`` if aperture is in the beam, otherwise returns ``False``.
        """

    def force_emit_signals(self) -> None:
        """Reemit all signals."""
        self.emit("valueChanged", self._current_position_name)
        self.emit(
            "diameterIndexChanged",
            self._current_diameter_index,
            self._diameter_size_list[self._current_diameter_index] / 1000.0,
        )



from mxcubecore.HardwareObjects import MiniDiff


class GrobDiff(MiniDiff.MiniDiff):
    def init(self):
        MiniDiff.MiniDiff.init(self)

        self.phiy_direction = -1

    def oscillate(self, range, exp_time, npasses=1):
        self.get_deviceby_role("phi").oscillation(range, exp_time, npasses)



import time

# from PyTango.gevent import DeviceProxy
from PyTango import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task


class TangoKeithleyPhotonFlux(HardwareObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        self.get_object_by_role("controller")
        self.shutter = self.get_deviceby_role("shutter")
        self.aperture = self.get_object_by_role("aperture")
        self.factor = self.get_property("current_photons_factor")

        self.shutter.connect("shutterStateChanged", self.shutterStateChanged)

        self.tg_device = DeviceProxy(self.get_property("tango_device"))
        self.counts_reading_task = self._read_counts_task(wait=False)

    @task
    def _read_counts_task(self):
        old_counts = None
        while True:
            counts = self._get_counts()
            if counts != old_counts:
                old_counts = counts
                self.countsUpdated(counts)
            time.sleep(1)

    def _get_counts(self):
        self.tg_device.MeasureSingle()
        counts = abs(self.tg_device.ReadData) * 1e6
        if self.aperture:
            try:
                aperture_coef = self.aperture.getApertureCoef()
            except Exception:
                aperture_coef = 1
        else:
            aperture_coef = 1

        counts *= aperture_coef
        return counts

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emitValueChanged()

    def shutterStateChanged(self, _):
        self.countsUpdated(self._get_counts())

    def updateFlux(self, _):
        self.countsUpdated(self._get_counts(), ignore_shutter_state=True)

    def countsUpdated(self, counts, ignore_shutter_state=False):
        if not ignore_shutter_state and self.shutter.getShutterState() != "opened":
            self.emitValueChanged(0)
            return
        flux = counts * self.factor
        self.emitValueChanged("%1.3g" % flux)

    def get_value(self):
        return self.current_flux

    def emitValueChanged(self, flux=None):
        self.current_flux = flux

        if flux is None:
            self.emit("valueChanged", ("?",))
        else:
            self.emit("valueChanged", (self.current_flux,))



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Control the detector cover

Example xml file:

.. code-block:: xml

  <object class="DetectorCover">
    <object href="/detcover" role="detector_cover"/>
    <object href="/handle_detcover" role="handle_detector_cover"/>
  </object>

.. code-block:: yaml

 class DetectorCover.DetectorCover
 configuration:
    username: Detector Cover
 objects:
    detector_cover: detcover.yaml
    handle_detector_cover: handle_detcover.yaml

"""

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class DetectorCover(HardwareObject):
    """Detector Cover class"""

    def __init__(self, name):
        super().__init__(name)
        self.detector_cover = None
        self.handle_detector_cover = True

    def init(self):
        """Initialise properties and objects"""
        super().init()
        self.detector_cover = self.get_object_by_role("detector_cover")
        self.handle_detector_cover = self.get_object_by_role("handle_detector_cover")

    def set_value(self, value, timeout: float | None = None):
        """Set the detector cover. Check if you need to handle it.
        Args:
            value: target value.
            timeout: optional - timeout [s].
                     If timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).
        """

        use = True
        if self.handle_detector_cover:
            try:
                use = self.handle_detector_cover.get_value().value
            except AttributeError:
                use = False
        if use:
            self.detector_cover.set_value(value, timeout)

    def get_value(self):
        """Get the values for the detector cover only."""
        return self.detector_cover.get_value()

    def get_value_handle(self):
        """Get the handle detector cover option."""
        if self.handle_detector_cover:
            return self.handle_detector_cover.get_value()
        return True



from mxcubecore.BaseHardwareObjects import HardwareObject


class ALBATransmission(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)
        self.transmission = None

    def init(self):
        self.transmissionChannel = self.get_channel_object("transmission")
        self.stateChannel = self.get_channel_object("state")

        self.transmissionChannel.connect_signal("update", self.transmissionChanged)
        self.stateChannel.connect_signal("update", self.stateChanged)

    def is_ready(self):
        return True

    def transmissionChanged(self, value):
        self.transmission = value
        self.emit("attFactorChanged", self.transmission)

    def stateChanged(self, value):
        self.state = str(value)
        self.emit("attStateChanged", self.state)

    def getAttState(self):
        self.state = self.stateChannel.get_value()
        return self.state

    def get_value(self):
        return self.transmissionChannel.get_value()

    def _set_value(self, value):
        self.transmissionChannel.set_value(value)

    def re_emit_values(self):
        value = self.get_value()
        self.emit("attFactorChanged", value)


def test_hwo(hwo):
    print("Transmission is: ", hwo.get_value())
